
/*
 * File     : atvme.c
 *
 * Purpose  : Driver for the library of access functions for Linux
 *            for AT_VME Revision 1.0, 2.0, ITVME6 und ITVME3
 *
 * DOS vers : (c) pro_VME Industrielle Rechnertechnik GmbH
 *            Postfach 1236, Wiesenbacherstr. 77
 *            D-6903 Neckargemuend (DOS library sources)
 *
 *
 * Last edit: 11.09.89 K.Hitschler
 *            20.09.90 K.Hitschler Anpassung an REL2.0
 *            27.11.91 R.Dux
 *
 *            31.07.98 H.-J. Mathes, begin porting to Linux
 *             1.10.98 D. Muehlenberg, continue the porting to Linux 
 *
 *             5.01.99 Th. Wiegert, changed sel_type from SEL_EX
 *                     to SEL_IN in atvme_select ()
 *
 *
 * $Revision: 0.24 $
 * $Date: 1999/02/18 10:42:03 $
 * $Id: atvme.c,v 0.24 1999/02/18 10:42:03 mathes Exp $
 */

#include <linux/unistd.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <linux/delay.h>

#include <asm/irq.h>
#include <asm/io.h>
#include <asm/mman.h>

#include <linux/atvme.h>

#define free kfree

#if 1
# define AtvmePrint(x)	{}
#else
# define AtvmePrint(x)	printk x;
#endif


#include "list.c"

/* --- some definitions for the source --- */

#undef DISABLE_INTS_ON_NEGLECT

/* Note: If this is defined !=0 the each read()/write() operation on
 *       the normal Vme bus address space is followed by reading the
 *       controllers error flag.
 *       If you are sure about what you are doing or if you want to have 
 *       a faster access you may set this to 0 !
 *       In any case, when reading interrupt related things, the
 *       error flag will be read.
 */

#define ATVME_DO_ERROR_CHECK	1

/* --- some necessary static definitions --- */
 
static unsigned int AtvmeMemBase = ATVME_CSR;

/* if TRUE, read()/write() will change the address set by lseek()
 * (i.e. fpos increment is tied to read()/write())
 */

static int Atvme_fpos_tied = TRUE;

static pid_t Atvme_s_pid = 0;

                                             /* base of port in i/O of AT    */
static unsigned int AtvmeIOBase = ATVME_IO;
                                             /*   shadow CSRegister          */
static unsigned short AtvmeCsrSave = ATVME_STATUS_DEFAULT;
                                             /* shadow aux CSRegister        */
static unsigned short AtvmeAuxCsrSave = ATVME_AUX_STATUS_DEFAULT;
                                             /* Default Interrupt-Level VME  */
static unsigned short AtvmeIntLevel = ATVME_LEVEL;

/* --- queue for the pending interrupts --- */

static List *pendingInts;

static unsigned int AtvmeIrq = ATVME_IRQ_LINE;

static int atvme_loaded = FALSE;
static int atvme_busy = FALSE;
static char atvme_driver_info[1024];

static AtvmeState atvme_state;

static IntAckStruct int_ack[0x100];

/*
 * definitions & prototypes for interface functions like <vme.h>/<vmetvc.h>
 */

static int atvmeReadA_D8 ( unsigned int );
static int atvmeReadA_D16 ( unsigned int );
static int atvmeReadA_D32 ( unsigned int );

static void atvmeWriteA_D8 ( unsigned int, int );
static void atvmeWriteA_D16 ( unsigned int, int );
static void atvmeWriteA_D32 ( unsigned int, int );

/* --- function tables used by atvme_read(), atvme_write() --- */

static int (*atvmeReadVTBL[])(unsigned int) = {
  atvmeReadA_D8,
  atvmeReadA_D16,
  atvmeReadA_D32
};

static void (*atvmeWriteVTBL[])(unsigned int, int) = {
  atvmeWriteA_D8,
  atvmeWriteA_D16,
  atvmeWriteA_D32
};

/* --- wait_queue used by interrupt handler and atvme_select() --- */

static struct wait_queue *wait_queue = NULL;

/* --- following kernel naming conventions (not mine) --- */

int  init_module ( void );
void cleanup_module ( void );
void atvme_init ( void );
void atvme_setup ( char *str, int *ints );

/* --- prototypes for methods accessing the hardware --- */

static int Atvme_check_io_ports ( void );

static void Atvme_init ( void );
static void Atvme_deinit ( void );
static int Atvme_check_connection ( void );
static void Atvme_reset ( void );
static int Atvme_set_am ( AtvmeAD *ad );
static unsigned int Atvme_get_addr_size ( void );
static int Atvme_test_addr_size ( unsigned int, int );
static IntAckStruct *Atvme_interrupt_ack ( void );
static void *Atvme_build_address ( unsigned int );
static void Atvme_free_irq ( int );
static void Atvme_rora_method(IntAckStruct *);

static void Atvme_clear_error ( void );

/* ------- DM 7.10.98 ------------------------------------------------------ */
/* ------- DM 12.11.98 ----------------------------------------------------- */

static void atvme_irq_handler(int irq, void *idontknow, struct pt_regs *ptr)
{
  int max; 
  IntAckStruct *user_int_ack;
                                    /* disable all interrupts */
  cli();
                                    /* init a int ack and return the vector */
  user_int_ack = Atvme_interrupt_ack();
                              /* add a IntAckStruct to the pendingInts-list */
  if ( user_int_ack )
   {
                             /* only if posted */
    addTail(pendingInts,user_int_ack);
    if ( ( max = getNumOfNodesInList(pendingInts) ) > ATVME_MAX_INTS_QUEUED )
     {
#ifdef DISABLE_INTS_ON_NEGLECT
      Atvme_free_irq ( AtvmeIrq );
      deleteList ( pendingInts );
      printk("atvme_irq_handler(): max. %d interrupts reached in queue. Disabling interrupts!\n",max);
#else
      resetList ( pendingInts );
      printk("atvme_irq_handler(): max. %d interrupts reached in queue. Erazing interrupts!\n",max);
#endif
    }
  }
                                    /* enable interrupts */
  sti();
  wake_up_interruptible ( &wait_queue );
  if ( Atvme_s_pid )
    kill_proc ( Atvme_s_pid, SIGUSR1, 0 );
}

/* --- registers in the CSR address space (memory) --- */

#define CSR            (AtvmeMemBase + 0x00)
#define DR             (AtvmeMemBase + 0x02)
#define ADR            (AtvmeMemBase + 0x04)
#define AMR            (AtvmeMemBase + 0x06)
#define ACSR           (AtvmeMemBase + 0x08)

/* ------------------------------------------------------------------------- */
/* ---      this stuff of the driver really accesses the hardware ...        */
/* ------------------------------------------------------------------------- */

#define Atvme_wait()	{ udelay ( 10 ); }

/*
 * Note: All functions which access the Atvme interface have
 *       to leave it in a state where the Vme bus is selected !!!
 */
 
#define Atvme_select_csr()        \
                outb(ATVME_OFF,AtvmeIOBase)    /* switch to CSR Space */
#define Atvme_select_vme()        \
                outb(ATVME_ON,AtvmeIOBase)     /* switch to VME Space */
#define Atvme_check_csr_mode()     \
                (inb(AtvmeIOBase)&1)     /* read back switch status */

#define Atvme_vme_on()            \
                outb(ATVME_ON,AtvmeIOBase+1)   /* enable interface    */
#define Atvme_vme_off()           \
                outb(ATVME_OFF,AtvmeIOBase+1)  /* disable interface */
#define Atvme_check_vme_mode()     \
                (inb(AtvmeIOBase+1)&1)   /* read back switch status */

#define Atvme_enable_int()        \
                outb(ATVME_ON,AtvmeIOBase+2)   /* enable extern Interrupts */
#define Atvme_disable_int()       \
                outb(ATVME_OFF,AtvmeIOBase+2)  /* disable extern Interrupts */
#define Atvme_get_int()           \
                (inb(AtvmeIOBase+2)&1)   /* fast VME Interrupt poll */

#define Atvme_check_error()    (inb(AtvmeIOBase+3)&1)   /* fast error read */

/* ------------------------------------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* ---                   interface to the file system                    --- */
/* ------------------------------------------------------------------------- */

static int atvme_lseek( struct inode * inode, struct file * file, 
                                 off_t offset, int origin );
static int atvme_read ( struct inode *inode, struct file *file, 
                                 char *buf,  int count );
static int atvme_write ( struct inode *inode, struct file *file, 
                                 const char *buf, int count );
static int atvme_select ( struct inode *inode, struct file *file, 
                                 int sel_type, select_table *wait );
static int atvme_ioctl ( struct inode *inode, struct file *file, 
                                 unsigned int cmd, unsigned long arg );
static int atvme_mmap ( struct inode *inode, struct file *file, 
                                 struct vm_area_struct * );
static int atvme_open ( struct inode *inode, struct file *file );
static void atvme_release ( struct inode *inode, struct file *file );

static struct file_operations atvme_fops =	/* VFS */
{
        atvme_lseek,    /* Seek */
        atvme_read,     /* Read */
        atvme_write,    /* Write */
        NULL,           /* Readdir */
        atvme_select,   /* Select */
        atvme_ioctl,    /* Ioctl */
        atvme_mmap,     /* MMap */
        atvme_open,     /* Open */
        atvme_release,  /* Release */
        NULL,
        NULL            /* Fasync */
};

/* ------------------------------------------------------------------------- */

/* called from: /usr/src/linux/drivers/char/mem.c */

void atvme_init ( void )
 {
  if ( pendingInts ) deleteList ( pendingInts );
  pendingInts = newList();

  init_module();
}

/* ------------------------------------------------------------------------- */

/* called from: /usr/src/linux/init/main.c */

void atvme_setup ( char *str, int *ints )
 {
  if ( ints[0] != 4 ) return;   /* not tested !!! */

  AtvmeIOBase = ints[1];
  AtvmeMemBase  = ints[2];
  AtvmeIrq = ints[3];
  AtvmeIntLevel = ints[4];
}

/* ------- DM 7.10.98 ------------------------------------------------------ */

int Atvme_enable_interrupt ( void )
 {
  int ret;

  ret = request_irq ( AtvmeIrq, atvme_irq_handler, 
                                SA_INTERRUPT, ATVME_NAME, NULL );
  if (ret)
   {
    switch (-ret)
    {
     case EINVAL : AtvmePrint (("invalid IRQ value %d (%s)\n",
                            AtvmeIrq, ATVME_NAME ));
                   return INIT_FAIL;

     case ENOMEM : AtvmePrint (("no memory for IRQ handler (IRQ %d, %s)\n",
                            AtvmeIrq, ATVME_NAME ));
                   return INIT_FAIL;

     case EBUSY  : printk ("%s : IRQ %d already busy\n",
                            ATVME_NAME, AtvmeIrq);
                   return INIT_FAIL;

     default     : printk ("could not install IRQ handler (IRQ %d, %s)\n",
                            AtvmeIrq, ATVME_NAME);
                   return INIT_FAIL;
    } 
/* this int is not available for the driver */
    AtvmeIrq = 0;
  } 
  else
   {

/* redirect VMEbus error or interrupts 
 * (TB or VME jumpered on ATADA) to PC-int AtvmeIrq 
 */
    Atvme_select_csr();

/* choose VME interrupts as int source */

    AtvmeCsrSave &= ATVME_ENABLE_VME_INT;
    writew(AtvmeCsrSave,CSR);
    Atvme_select_vme();

    printk ( "%s: Handler for IRQ %d successfully installed.\n",
             ATVME_NAME, AtvmeIrq );
    Atvme_enable_int();

    enable_irq ( AtvmeIrq );
  }
  
  return ret;
}

/* ------------------------------------------------------------------------- */

int init_module ( void )
 {
  int io_check, connect;
  char connect_str[30];

  if ( register_chrdev ( ATVME_MAJOR, ATVME_NAME, &atvme_fops ) ) 
   {
    printk ( "unable to get major %i for %s device\n", 
             ATVME_MAJOR, ATVME_NAME );
    return INIT_FAIL;
  }

  printk ( "Major %i for %s device has been successfully registered.\n", 
           ATVME_MAJOR, ATVME_NAME );
 
  if ( check_region ( AtvmeIOBase, ATVME_IO_LENGTH ) ) 
   {
    printk ( " unable to access i/o address 0x%04X\n", AtvmeIOBase );
    unregister_chrdev ( ATVME_MAJOR, ATVME_NAME );
    return INIT_FAIL;
  }

  printk ( "%s: Ports 0x%04X - 0x%04X have been reserved.\n",
          ATVME_NAME, AtvmeIOBase, AtvmeIOBase + ATVME_IO_LENGTH - 1 );

  request_region ( AtvmeIOBase, ATVME_IO_LENGTH, ATVME_NAME);

  atvme_loaded = TRUE;

  io_check = Atvme_check_io_ports();
  if ( io_check ) printk ( "%s found in IO space!\n", ATVME_NAME );
  
  Atvme_init();

  connect = Atvme_check_connection();

  switch ( connect )
   {
    case 3: strcpy ( connect_str, "Crate HE3 connected" );
            break;
    case 6: strcpy ( connect_str, "Crate HE6 connected" );
            break;
    default:
            strcpy ( connect_str, "No crate connected" );
  }

  Atvme_deinit();
  
  sprintf ( atvme_driver_info, 
            "%s driver initialized with \nIO: 0x%x  Mem: 0x%x and %s\n%s\n",
            ATVME_NAME, AtvmeIOBase, AtvmeMemBase, connect_str,
	    "$Revision: 0.24 $ $Date: 1999/02/18 10:42:03 $" );

  AtvmePrint (( atvme_driver_info ));
  
  return INIT_SUCCESS;
}

/* ------ DM 18.11.98 ------------------------------------------------------- */

void Atvme_reset_int_ack(void)
 {
  int i;

  for ( i=0; i<NUMBER(int_ack); i++ )
   {
    int_ack[i].int_vec = -1;
  }
}

/* ------ DM 9.10.98 ------------------------------------------------------- */

void Atvme_free_irq ( int irq )
 {
  if ( AtvmeIrq > 0 )
   {
    disable_irq ( AtvmeIrq );
    free_irq ( AtvmeIrq, NULL );
  }
}

/* ------------------------------------------------------------------------- */

void cleanup_module ( void )
 {
  if (MOD_IN_USE)
   {
    printk ( "%s: device is busy\n", ATVME_NAME );
    return;
  }

  Atvme_deinit();
  
  release_region ( AtvmeIOBase, ATVME_IO_LENGTH );
 
  Atvme_free_irq ( AtvmeIrq );
  
  if ( unregister_chrdev ( ATVME_MAJOR, ATVME_NAME ) != 0 )
    printk ( "%s: cleanup_module failed\n", ATVME_NAME );
  else
    printk ( "%s: module removed\n", ATVME_NAME );
}

/* ------------------------------------------------------------------------- */

static int atvme_open ( struct inode *inode, struct file *file )
 {
  MOD_INC_USE_COUNT;

  if ( atvme_busy ) 
   {
    MOD_DEC_USE_COUNT;
    printk ("%s: device already in use\n", ATVME_NAME);
    return -EADDRINUSE;
  }
  
  atvme_busy = TRUE;
  
  Atvme_init();

  return 0;
}

/* ------- DM 1.10.98 ------------------------------------------------------ */

static unsigned short Atvme_build_am ( unsigned short addr_type, 
                                       unsigned short access_type )
 {
  int in, priv;
  static unsigned short ams[] = {
  /* 16 bit address */
    AM_SHORT_SUP,
    AM_SHORT_SUP,
    AM_SHORT_SUP,
    AM_SHORT_NONPRIV,
    AM_SHORT_NONPRIV,
    AM_SHORT_NONPRIV,

  /* 24 bit address */
    AM_STD_SUP_BLOCK,
    AM_STD_SUP_PROG,
    AM_STD_SUP_DATA,
    AM_STD_NONPRIV_BLOCK,
    AM_STD_NONPRIV_PROG,
    AM_STD_NONPRIV_DATA,

  /* 32 bit address */
    AM_EXT_SUP_BLOCK,
    AM_EXT_SUP_PROG,
    AM_EXT_SUP_DATA,
    AM_EXT_NONPRIV_BLOCK,
    AM_EXT_NONPRIV_PROG,
    AM_EXT_NONPRIV_DATA,

  /* 64 bit address */
    AM_LONG_SUP_BLOCK,
    AM_LONG_SUP_PROG,
    AM_LONG_SUP_DATA,
    AM_LONG_NONPRIV_BLOCK,
    AM_LONG_NONPRIV_PROG,
    AM_LONG_NONPRIV_DATA
  };
 
  priv = (access_type & ATVME_PRIV_MASK) >> 8;
  access_type = (access_type & ATVME_ACC_MASK) >> 12;
  addr_type &= ATVME_A_MASK;
  in = (addr_type-1)*6;
  in += (priv-1)*3;
  in += access_type-1;
  if (in < 0 || in >= NUMBER(ams)) return 0;
 
  return ams[in];
}

/* ---------- DM 2.10.98 -------------------------------------------------- */

static int Atvme_set_am ( AtvmeAD *ad )
 {
  int am;
  int addr_type, acc_type;

  if ( !ad ) return FAIL;

  if (
      atvme_state.ad.addr_type == ad->addr_type &&
      atvme_state.ad.data_type == ad->data_type &&
      atvme_state.ad.acc_type == ad->acc_type
  ) return SUCCESS;

  addr_type = ad->addr_type & ATVME_A_MASK;
  acc_type = ad->acc_type & ( ATVME_ACC_MASK | ATVME_PRIV_MASK );

  if ( !addr_type || !acc_type )
   {
    return FAIL;
  }
  
  am = Atvme_build_am ( addr_type, acc_type );
  AtvmePrint(( "Atvme_set_am(): am 0x%x\n", am ));
  
  if ( am < 0 ) return FAIL;

  atvme_state.am = am;
  atvme_state.ad.addr_type = ad->addr_type;
  atvme_state.ad.data_type = ad->data_type;
  atvme_state.ad.acc_type = ad->acc_type;

  Atvme_select_csr();
  writew ( am, AMR );

  if ( atvme_state.ad.data_type == ATVME_D32 )
   {
    AtvmeCsrSave &= ATVME_ENABLE_LWORD;
  }
  else
   {
    AtvmeCsrSave |= ATVME_DISABLE_LWORD;
  } 
  writew ( AtvmeCsrSave, CSR );
  
  Atvme_select_vme();

  return SUCCESS;
}

/* ---------- DM 2.10.98 -------------------------------------------------- */

static inline unsigned int Atvme_get_addr_size ( void )
 {
  if (atvme_state.ad.addr_type == 1)
   {
    return 0xFFFF;
  }
  else if (atvme_state.ad.addr_type == 2)
   {
    return 0xFFFFFF;
  }
  else
   {
    return 0xFFFFFFFF;
  } 
}

/* ---------- DM 2.10.98 -------------------------------------------------- */

static inline int Atvme_test_addr_size ( unsigned int fpos, int offset )
 {
  if ( fpos + offset < 0 ) return FALSE;
  if ( fpos + offset >= Atvme_get_addr_size() ) return FALSE;

  return TRUE;
}

/* ---------- DM 2.10.98 -------------------------------------------------- */

static int atvme_lseek ( struct inode * inode, struct file * file, 
                                               off_t offset, int origin)
 {
  switch ( origin )
   {
    case 0:	/* SEEK_SET */
      if ( !Atvme_test_addr_size ( 0, offset ) )
       {
        return -EINVAL;
      }
      file->f_pos = offset;
      AtvmePrint (( "atvme_lseek(SET): f_pos 0x%08X\n", (int)file->f_pos ));
    break;

    case 1:	/* SEEK_CUR */
      if ( !Atvme_test_addr_size ( file->f_pos, offset ) )
       {
        return -EINVAL;
      }
      file->f_pos += offset;
      AtvmePrint (( "atvme_lseek(CUR): f_pos 0x%08X\n", (int)file->f_pos ));
    break;

    case 2:	/* SEEK_END */
      if ( offset >= 0 )
       {
        return -EINVAL;
      }
      file->f_pos =  Atvme_get_addr_size() + offset;
      AtvmePrint (( "atvme_lseek(END): f_pos 0x%08X\n", (int)file->f_pos ));
    break;
  }

  return file->f_pos;
}

/* ------------------------------------------------------------------------- */

static int atvme_ioctl ( struct inode* inode, struct file* file, 
                         unsigned int cmd, unsigned long arg )

 {
  int ret,vec;
  AtvmeAD ad;
  IntAckStruct user_int_ack;
  Node *node;
  
  switch ( cmd )
   {
    case ATVME_INITIALIZE        : 
         Atvme_init();
         return 0;
    
    case ATVME_SYSRESET          :
         Atvme_reset();
         return 0;

    case ATVME_CHECK_CONNECTION  :
         ret = Atvme_check_connection();
         put_fs_long ( (unsigned long)ret, (unsigned long *)arg );
         return 0;

    case ATVME_SET_AM:
         memcpy_fromfs ( &ad, (AtvmeAD *)arg, sizeof(ad) );
	 if ( !Atvme_set_am ( (AtvmeAD *)&ad ) )
	  {
	   printk ( "ioctl()::ATVME_SET_AM: no AM addr %x, data %x, acc %x\n", 
	            ad.addr_type, ad.data_type, ad.acc_type );
	   return -EINVAL;
	 }
         return 0;  

    case ATVME_ENABLE_INTERRUPT:
         Atvme_enable_interrupt();
         return 0;

    case ATVME_DISABLE_INTERRUPT:
         Atvme_free_irq ( AtvmeIrq );
         return 0;
	
    case ATVME_POST_INTERRUPT:
         memcpy_fromfs(&user_int_ack, (IntAckStruct *)arg, sizeof(IntAckStruct));
         vec = user_int_ack.int_vec; 
         if (vec < 0 || vec >= NUMBER(int_ack)) {
           printk("ioctl()::ATVME_POST_INTERRUPT: bad int vec %x\n!",vec);
           return -EINVAL;
         }
         memcpy(int_ack+vec,&user_int_ack,sizeof(IntAckStruct));
         return 0;

    case ATVME_REMOVE_INTERRUPT:
         vec = (int) arg;
         if (vec < 0 || vec >= NUMBER(int_ack)) return -EINVAL;
         int_ack[vec].int_vec = -1;
         return 0;

    case ATVME_GET_INTERRUPT:
         cli();
         node = getFirstNode ( pendingInts );
         if (!node) {
           sti();
           return -EINVAL;
         }
         memcpy_tofs ( (void *)arg, node->data, sizeof(IntAckStruct) );
         delFirstNodeAndData ( pendingInts );
         sti();
         return 0;

    case ATVME_ATTACH_SIGNAL:
         Atvme_s_pid = (int) arg;
         return 0;

    case ATVME_SET_64K_PAGE:
         Atvme_build_address ( (unsigned int)arg );
         return 0;

    case ATVME_GET_DRIVER_INFO    : 
         memcpy_tofs ( (void*)arg, atvme_driver_info, 
                                  sizeof ( atvme_driver_info ) );
         return 0;

    case ATVME_ERROR_CHECK:
         ret = Atvme_check_error();
         if (ret) {
           Atvme_clear_error();
         }
         memcpy_tofs((int *)arg,&ret,sizeof(ret));
         return 0;

    case ATVME_TIE_FPOS:
         Atvme_fpos_tied = TRUE;
	 AtvmePrint(( "Atvme_fpos_tied: %d\n", Atvme_fpos_tied ));
         return 0;

    case ATVME_UNTIE_FPOS:
         Atvme_fpos_tied = FALSE;
	 AtvmePrint(( "Atvme_fpos_tied: %d\n", Atvme_fpos_tied ));
         return 0;
  }

  return ( -EINVAL );
}

/* ------------------------------------------------------------------------- */

static void atvme_release ( struct inode *inode, struct file *file )
 {
  Atvme_deinit();
  
  if ( AtvmeIrq )
   {
    disable_irq ( AtvmeIrq );
    free_irq ( AtvmeIrq, NULL );
  }
 
  atvme_busy = FALSE; 
  MOD_DEC_USE_COUNT;
}

/* ------ DM 2.10.98 ------------------------------------------------------- */

static void *Atvme_build_address ( unsigned int offs )
 {
  unsigned short msw;

  switch ( atvme_state.ad.addr_type )
   {
    case ATVME_A16:
      Atvme_select_vme();
      return (void *)( AtvmeMemBase + offs );
    break;
  
    case ATVME_A24:
      msw = ( offs & 0x00ff0000 ) >> 16;
      if ( atvme_state.msw_addr != msw )
       {
        atvme_state.msw_addr = msw;
        Atvme_select_csr();
        writew ( msw, ADR );
        Atvme_select_vme();
      }
      return (void *)( AtvmeMemBase + ( offs & 0xffff ) );
    break;
  
    case ATVME_A32: 
      msw = ( offs & 0xffff0000 ) >> 16;
      if ( atvme_state.msw_addr != msw )
       {
        atvme_state.msw_addr = msw;
        Atvme_select_csr();
        writew ( msw, ADR );
        Atvme_select_vme();
      } 
      return (void *)( AtvmeMemBase + ( offs & 0xffff ) );
    break;
  
    default: return NULL;
  }
}

/* ------ DM 2.10.98 ------------------------------------------------------- */
/* ------ DM 11.11.98 ------------------------------------------------------ */

static void atvmeWriteA_D8 ( unsigned int addr, int data )
 {
  char *ptr = Atvme_build_address ( addr );

  if ( !ptr ) return;
  writeb ( data, ptr );
}

static void atvmeWriteA_D16 ( unsigned int addr, int data )
 {
  char *ptr = Atvme_build_address ( addr );

  if ( !ptr ) return;
  writew(data,ptr);
}

static void atvmeWriteA_D32 ( unsigned int addr, int data )
 {
  char *ptr;

  ptr = Atvme_build_address ( addr );
  if ( !ptr ) return;

  Atvme_select_csr();
  writew( (unsigned short)( data >> 16 ), DR );
  Atvme_select_vme();

  writew ( data & 0xFFFF, ptr );
}

static int atvmeReadA_D8 ( unsigned int addr )
 {
  char *ptr = Atvme_build_address ( addr );

  if ( !ptr ) return 0x7f;
  return readb ( ptr );
}

static int atvmeReadA_D16 ( unsigned int addr )
 {
  char *ptr = Atvme_build_address ( addr );

  if ( !ptr ) return 0x7fff;
  return readw ( ptr );
}

static int atvmeReadA_D32 ( unsigned int addr )
 {
  char *ptr;
  int data;

  ptr = Atvme_build_address ( addr );
  if ( !ptr ) return 0x7ffffff;

  data = readw ( ptr );
   
  Atvme_select_csr();
  data |= readw ( DR ) << 16;
  Atvme_select_vme();
  
  return data;
}

/* ------ DM 6.10.98 ------------------------------------------------------- */

static int atvme_select ( struct inode *inode, struct file *file, 
                          int sel_type, select_table *wait )
 {
  if ( sel_type != SEL_IN ) return 1;
  if ( !emptyList ( pendingInts ) ) return 1;
  if ( !Atvme_get_int() )
   {
    select_wait ( &wait_queue, wait );
    return 0;
  }
  
  return 1;
}

/* ------ DM 5.10.98 ------------------------------------------------------- */

static int atvme_mmap ( struct inode *inode, struct file *file, 
                                             struct vm_area_struct *vma )
{
  if ( vma->vm_offset & ~PAGE_MASK )
   {
    printk ( "atvme_mmap(): unaligned access %ld\n", vma->vm_offset );
    return -ENXIO;
  }

  if ( remap_page_range ( vma->vm_start, vma->vm_offset, 
              vma->vm_end - vma->vm_start, vma->vm_page_prot ) )
   {
    printk ( "atvme_mmap(): can't remap_page_range!\n" );
    return -EAGAIN;
  }
  
  vma->vm_inode = inode;
  inode->i_count++;

  return 0;
}


/* ------ DM 26.1.98 ------------------------------------------------------- */

/* this is to use the macros __put_user() and __get_user() */

int bad_user_access_length(void)
 {
  printk("bad_user_access_length not executed!\n");

  return 0;
}

/* ------ DM 2.10.98 ------------------------------------------------------- */

static int atvme_read ( struct inode *inode, struct file *file, 
                                             char *buf, int count )
 {
  int i,w = 0;
  int dtype=1;
  int data_type = ((atvme_state.ad.data_type & ATVME_D_MASK)>>4)-1;
  unsigned int end = Atvme_get_addr_size();

  /* --- if you try to read past end --- */

  if ( file->f_pos + count >= end )
   {
    AtvmePrint (( "atvme_read(): file->f_pos 0x%08X, end 0x%08X!\n",
                  (int)file->f_pos, end ));
    count = end - file->f_pos;
  }

  dtype = 1<<data_type;

  for ( i=0; i<count; i+=dtype )
   {
    AtvmePrint (( "atvme_read(): f_pos = 0x%08X\n", (int)(file->f_pos) ));
    w = atvmeReadVTBL[data_type](file->f_pos);

#if ( ATVME_DO_ERROR_CHECK == 1 )
    if ( Atvme_check_error() )
     {
      printk ( "atvme_read(): error occured during vmebus access!\n" );
      Atvme_clear_error();

      return i;
    }
#endif

/* this is really funny:           */
/* in the function __put_user(),   */
/* __get_user() is a call to       */
/* bad_user_access_length(), which */
/* is not exported to modules!!!   */
/* so you are only able to use     */
/* these funcs with constant size  */
/* parameters i.e. in defines,     */
/* else the compiler won't cut     */
/* the default statement and so    */
/* the linker is looking for       */
/* bad_user..(). It's really nice. */
/* It's very stupid, 'cos you have */
/* to go through the same switch-  */
/* statement twice!!!              */ 
#if 1  
    __put_user(w,buf+i,dtype);
    /* can't use this! */
#else
    switch ( dtype )
     {
      case 1:
        put_fs_byte ( w, buf+i );
      break;

      case 2:
        put_fs_word ( w, buf+i );
      break;

      case 4:
        put_fs_long ( w, buf+i );
      break;

      default: 
      break;
    }
#endif

    if ( Atvme_fpos_tied ) file->f_pos += dtype;
  }  /* for ( i=.. ) */

  return i;
}

/* ------ DM 2.10.98 ------------------------------------------------------- */

static int atvme_write ( struct inode *inode, 
                         struct file *file, const char *buf, int count )
 {
  int i,w,dtype=1;
  int data_type = ((atvme_state.ad.data_type & ATVME_D_MASK)>>4)-1;
  unsigned int end = Atvme_get_addr_size();

  /* --- if you try to read past end --- */

  if ( file->f_pos + count >= end )
   {
    count = end - file->f_pos;
  }

  dtype = 1<<data_type;

  for ( i=0; i<count; i+=dtype )
   {
#if 1  
    w = (int)__get_user ( buf+i, dtype );
#else
    switch ( dtype )
     {
      case 1:
        w = get_fs_byte ( buf+i );
      break;

      case 2:
        w = get_fs_word ( buf+i );
      break;

      case 4:
        w = get_fs_long ( buf+i );
      break;

      default: w = 0;
      break;
    }
#endif

    atvmeWriteVTBL[data_type]( file->f_pos, w );
    AtvmePrint (( "atvme_write(): f_pos = 0x%08X\n", (int)(file->f_pos) ));

#if ( ATVME_DO_ERROR_CHECK == 1 )
    if ( Atvme_check_error() )
     {
      AtvmePrint (( "atvme_write(): error occured during vmebus access!\n" ));
      Atvme_clear_error();

      return i;
    }
#endif
    
    if ( Atvme_fpos_tied ) file->f_pos += dtype;
  }

  return i;
}

/* ------------------------------------------------------------------------- */

/* 
 * perform short read/write test on registers in I/O space ...
 */
 
static int Atvme_check_io_ports ( void )
 {
  int found = 0;
  
  outb_p ( ATVME_ON, AtvmeIOBase );
  if ( ( inb_p ( AtvmeIOBase ) & ATVME_ON ) == ATVME_ON ) found++;
  outb_p ( ATVME_ON, AtvmeIOBase+1 );
  if ( ( inb_p ( AtvmeIOBase+1 ) & ATVME_ON ) == ATVME_ON ) found++;
  outb_p ( ATVME_ON, AtvmeIOBase+2 );
  if ( ( inb_p ( AtvmeIOBase+2 ) & ATVME_ON ) == ATVME_ON ) found++;
  
  return ( found );
}

/* ------------------------------------------------------------------------- */

static void Atvme_init ( void )
 {
  if ( pendingInts ) deleteList ( pendingInts );
  pendingInts = newList();

  Atvme_reset_int_ack(); 
  Atvme_fpos_tied = TRUE;
  Atvme_s_pid = 0;

  Atvme_vme_on();                                 /* Switch interface ATVME_ON */
  Atvme_select_csr();                             /* Switch to Status */
  
#ifdef ATVME_REL2_0
  AtvmeAuxCsrSave = ATVME_AUX_STATUS_DEFAULT;
  writew ( AtvmeAuxCsrSave, ACSR );               /* clear SysFail */
#endif

  /* init AtvmeState ... */
  
  atvme_state.ad.addr_type = ATVME_A16;
  atvme_state.ad.data_type = ATVME_D16;
  atvme_state.ad.acc_type = ATVME_NONPRIV | ATVME_DATA;
  atvme_state.msw_addr = ATVME_HI_ADR_DEFAULT;
  atvme_state.am = ATVME_AM_DEFAULT;

  writew ( 0xffff, CSR );                         /* Disable Error */
  AtvmeCsrSave  = ATVME_STATUS_DEFAULT;
  writew ( AtvmeCsrSave, CSR );                   /* Enable Error and Word */
  
  writew ( atvme_state.am, AMR );                 /* Write Address Modifier */
  writew ( atvme_state.msw_addr, ADR );           /* Set current HighAddress */
  writew ( ATVME_HI_DATA_DEFAULT, DR );           /* Set current High Data */

  Atvme_select_vme();                             /* Switch to Mem */
  if ( Atvme_check_error() ) {
    Atvme_clear_error();                          /* Clear Error if no MB  */
  }

  Atvme_wait();                                   /* Settling time needed  */
}

/* ------------------------------------------------------------------------- */

static void Atvme_deinit ( void )
 {
  Atvme_vme_off();                                                 /* do it */
}

/* ------------------------------------------------------------------------- */

static int Atvme_check_connection ( void )
 {
  int ret = 0;

  Atvme_select_csr();
  writew ( 0xffff, CSR );                        /* set all to inactive mode */
  Atvme_wait();
  if ( ( readw(CSR) & 0x001f ) != 0x001f ) return(0);       /* test */

  writew ( ATVME_ENABLE_LWORD, CSR );                 /* toggle bit and test */
  Atvme_wait();
  if ( !( ~readw(CSR) & ATVME_DISABLE_LWORD ) )
    ret = 3;
  else 
    ret = 6;

  writew ( ATVME_ENABLE_ERROR, CSR );
  Atvme_wait();
  if ( !( ~readw(CSR) & ATVME_DISABLE_ERROR ) ) return(0);

  writew ( ATVME_READ_INT_VECTOR, CSR );
  Atvme_wait();
  if ( !( ~readw(CSR) & ATVME_DISABLE_IACK ) ) return(0);

  writew ( AtvmeCsrSave, CSR );                       /* restore all settings */
  Atvme_select_vme();

  return ( ret );
}

/* ------------------------------------------------------------------------- */

static void Atvme_reset ( void )
 {
  AtvmeAuxCsrSave |= ATVME_SYSRES_SET;
  Atvme_select_csr();                                    /* switch to Status */
  writew ( AtvmeAuxCsrSave, ACSR );                      /* Sysreset Vme     */
  Atvme_wait();                                          /* wait           */
  AtvmeAuxCsrSave &= ( ATVME_SYSRES_CLEAR & ATVME_SYSFAIL_CLEAR & 
                       ATVME_SYSRES_INT_CLEAR ); /* auch REL 2.0  */
  writew ( AtvmeAuxCsrSave, ACSR );
  Atvme_wait();
                   /* Reset und Sysfail und Resetinterrupt ruecksetzen */
  Atvme_select_vme();                                          /* Switch to Mem */
}

/* ------------------------------------------------------------------------- */

void Atvme_clear_error ( void )
                                                    /* clears the ERROR flag */
 {
  Atvme_select_csr();                                    /* Switch to Status */
                                                 /* Write Status, dis. error */
  writew ( AtvmeCsrSave | ATVME_DISABLE_ERROR, CSR );
  writew ( AtvmeCsrSave, CSR );                /* Write Status, enable error */
  
  Atvme_select_vme();                                       /* Switch to Mem */
}

/* ------ DM 20.11.98 ------------------------------------------------------- */

void Atvme_rora_method ( IntAckStruct *ia )
 {
  AtvmeAD old_ad = atvme_state.ad,ad;
  int data_type;

  AtvmePrint (( "Atvme_rora_method(): register = 0x%x\n", ia->rora_reg ));

  ad.addr_type = ia->acc_mask & ATVME_A_MASK;
  ad.data_type = ia->acc_mask & ATVME_D_MASK;
  ad.acc_type = ia->acc_mask & ( ATVME_ACC_MASK | ATVME_PRIV_MASK );

  Atvme_set_am ( &ad );
  data_type = ((ad.data_type & ATVME_D_MASK)>>4)-1;

  atvmeWriteVTBL[data_type]( ia->rora_reg, 1 );

  if ( Atvme_check_error() )
   {
    Atvme_clear_error();
    atvmeReadVTBL[data_type](ia->rora_reg);
    if ( Atvme_check_error() )
     {
      printk("atvme_rora_method(): error occured during register access!\n");
      Atvme_clear_error();
    }
  }

  Atvme_set_am ( &old_ad );
}

/* ------ DM 7.10.98 ------------------------------------------------------- */
/* ------ DM 20.11.98 ------------------------------------------------------ */

static IntAckStruct *Atvme_interrupt_ack ( void )
 {
  IntAckStruct *user_int_ack = NULL;
  void *addr;
  int int_vec;

  addr = Atvme_build_address( (AtvmeIntLevel << 1) & 0x0e );

  Atvme_select_csr();
  writew ( AtvmeCsrSave & ATVME_READ_INT_VECTOR, CSR );
  Atvme_select_vme();

  int_vec = readw ( addr ) & 0xFF;
  if ( Atvme_check_error() )
   {
    Atvme_clear_error();
    printk ( "Atvme_interrupt_ack(): error occured during register access!\n" );
  }

  AtvmePrint (( "Atvme_interrupt_ack(): int_vec = %x\n", int_vec ));
  
  if (int_vec == int_ack[int_vec].int_vec)
   {
    user_int_ack = NEW_GFP_ATOMIC(IntAckStruct);
    if ( !user_int_ack )
     {
      printk("Atvme_interrupt_ack(): no memory for IntAckStruct!\n");
      Atvme_select_csr();
      writew ( AtvmeCsrSave, CSR );
      Atvme_select_vme();
      return NULL;
    } 
    memcpy ( user_int_ack, int_ack+int_vec, sizeof(IntAckStruct) );
    user_int_ack->int_tl = jiffies;
    if ( int_ack[int_vec].ack_type == AtvmeRoraByHandler )
     {
      Atvme_rora_method ( int_ack + int_vec );
    }
  }
  else
   {
    printk("Atvme_interrupt_ack(): int_vec %x not registered for this handler!\n",int_vec);
  }
    
  Atvme_select_csr();
  writew ( AtvmeCsrSave, CSR );
  Atvme_select_vme();

  return user_int_ack;
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
